#include "util/mongodbexperimentscope.h"

#include "util/conversions.h"

#include "mongocxx/collection.hpp"
#include "mongocxx/pipeline.hpp"

#include "bsoncxx/builder/stream/document.hpp"
#include "bsoncxx/oid.hpp"

#include <iostream>

namespace satlab
{
 
MongoDBExperimentScope::InstanceIncrementer::InstanceIncrementer(
    mongocxx::cursor::iterator&& cursorIt)
: cursorIt(std::move(cursorIt))
{
}

void MongoDBExperimentScope::InstanceIncrementer::next()
{
    BASE::InstanceIncrementer::next();
    
    cursorIt++;
    
    instanceIsUpdated = false;
}

void MongoDBExperimentScope::InstanceIncrementer::updateInstance()
{
    if(instanceIsUpdated)
    {
        return;
    }
    
    instance = KSATinstance(types::convert(*cursorIt));
    
    instanceIsUpdated = true;
}

KSATinstance* MongoDBExperimentScope::InstanceIncrementer::operator->()
{
    updateInstance();
    
    return &instance;
}

KSATinstance& MongoDBExperimentScope::InstanceIncrementer::operator*()
{
    updateInstance();
    
    return instance;
}


MongoDBExperimentScope::MongoDBExperimentScope(mongocxx::cursor&& cursor)
: cursor(std::move(cursor))
{
}

MongoDBExperimentScope::MongoDBExperimentScope(const mongocxx::database& db, 
                                               const std::string& scopeName)
: MongoDBExperimentScope(std::move(getCursor(db, scopeName)))
{    
}

mongocxx::cursor MongoDBExperimentScope::getCursor(const mongocxx::database& db,
                                                   const std::string& scopeName)
{
    return db["experiment_scopes"].aggregate(createScopePipeline(scopeName));
}

mongocxx::pipeline 
MongoDBExperimentScope::createScopePipeline(const std::string& scopeName)
{
    mongocxx::pipeline pipeline;
    
    pipeline.match(bsoncxx::builder::stream::document()
                   << "_id" << scopeName
                   << bsoncxx::builder::stream::finalize);
    
    pipeline.unwind("$instances");
    
    pipeline.lookup(bsoncxx::builder::stream::document()
                    << "from" << "instances"
                    << "localField" << "instances"
                    << "foreignField" << "_id"
                    << "as" << "instance"
                    << bsoncxx::builder::stream::finalize);
    
    pipeline.unwind("$instance");
    
    pipeline.replace_root(bsoncxx::builder::stream::document()
                          << "newRoot" << "$instance"
                          << bsoncxx::builder::stream::finalize);
    
    return pipeline;
}

MongoDBExperimentScope::BASE::iterator MongoDBExperimentScope::begin() noexcept
{
    P_INCR_T incrementer = P_INCR_T(new InstanceIncrementer(cursor.begin()));
    
    return BASE::createIterator(
        std::move(incrementer),
                                
        [] (const std::shared_ptr<BASE::InstanceIncrementer>& lhs,
            const std::shared_ptr<BASE::InstanceIncrementer>& rhs)
        {
            auto derivedLhs = std::dynamic_pointer_cast<InstanceIncrementer>(lhs);
            auto derivedRhs = std::dynamic_pointer_cast<InstanceIncrementer>(rhs);
            
            if(derivedLhs && derivedRhs)
            {
                return derivedLhs->cursorIt == derivedRhs->cursorIt;
            }
        
            return false;
        }
    );
}

MongoDBExperimentScope::BASE::iterator MongoDBExperimentScope::end() noexcept
{    
    P_INCR_T incrementer = P_INCR_T(new InstanceIncrementer(cursor.end()));
    
    return BASE::createIterator(
        std::move(incrementer),
                                
        [] (const std::shared_ptr<BASE::InstanceIncrementer>& lhs,
            const std::shared_ptr<BASE::InstanceIncrementer>& rhs)
        {
            auto derivedLhs = std::dynamic_pointer_cast<InstanceIncrementer>(lhs);
            auto derivedRhs = std::dynamic_pointer_cast<InstanceIncrementer>(rhs);
            
            if(derivedLhs && derivedRhs)
            {
                return derivedLhs->cursorIt == derivedRhs->cursorIt;
            }
            return false;
        }
    );
}

}
